#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utility functions for logging."""

import logging
import os
import threading
from contextvars import ContextVar
from datetime import datetime
from types import TracebackType
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union, cast
from uuid import UUID, uuid4

from pydantic import BaseModel, ConfigDict, Field

from zenml.client import Client
from zenml.config.pipeline_configurations import PipelineConfiguration
from zenml.config.step_configurations import StepConfiguration
from zenml.constants import (
    ENV_ZENML_DISABLE_PIPELINE_LOGS_STORAGE,
    ENV_ZENML_DISABLE_STEP_LOGS_STORAGE,
    ENV_ZENML_SERVER,
    handle_bool_env_var,
)
from zenml.enums import LoggingLevels, StackComponentType
from zenml.exceptions import DoesNotExistException
from zenml.logger import get_logger
from zenml.models import (
    LogsRequest,
    LogsResponse,
    LogsUpdate,
    PipelineRunResponse,
    StepRunResponse,
)
from zenml.utils import context_utils

if TYPE_CHECKING:
    from zenml.log_stores.base_log_store import BaseLogStoreOrigin
    from zenml.zen_stores.base_zen_store import BaseZenStore

logger = get_logger(__name__)


class LogEntry(BaseModel):
    """A structured log entry with parsed information.

    This is used in two distinct ways:
        1. If we are using the artifact log store, we save the
        entries as JSON-serialized LogEntry's in the artifact store.
        2. When queried, the server returns logs as a list of LogEntry's.
    """

    message: str = Field(description="The log message content")
    name: Optional[str] = Field(
        default=None,
        description="The name of the logger",
    )
    level: Optional[LoggingLevels] = Field(
        default=None,
        description="The log level",
    )
    timestamp: Optional[datetime] = Field(
        default=None,
        description="When the log was created",
    )
    module: Optional[str] = Field(
        default=None, description="The module that generated this log entry"
    )
    filename: Optional[str] = Field(
        default=None,
        description="The name of the file that generated this log entry",
    )
    lineno: Optional[int] = Field(
        default=None, description="The fileno that generated this log entry"
    )
    chunk_index: int = Field(
        default=0,
        description="The index of the chunk in the log entry",
    )
    total_chunks: int = Field(
        default=1,
        description="The total number of chunks in the log entry",
    )
    id: UUID = Field(
        default_factory=uuid4,
        description="The unique identifier of the log entry",
    )

    model_config = ConfigDict(
        # ignore extra attributes during model initialization
        extra="ignore",
    )


class LoggingContext(context_utils.BaseContext):
    """Context manager which collects logs using a LogStore."""

    __context_var__ = ContextVar("logging_context")

    def __init__(
        self,
        name: str,
        log_model: "LogsResponse",
        block_on_exit: bool = True,
        **metadata: Any,
    ) -> None:
        """Initialize the logging context.

        Args:
            name: The name of the logging context.
            log_model: The logs response model for this context.
            block_on_exit: Whether to block until all logs are flushed when the
                context is exited, if there are no more logging contexts active.
            **metadata: Additional metadata to attach to the log entry.
        """
        self.log_model = log_model
        self._lock = threading.Lock()
        self._disabled = False
        self._log_store = Client().active_stack.log_store
        self._metadata = metadata
        self._origin: Optional["BaseLogStoreOrigin"] = None
        self._name = name
        self._block_on_exit = block_on_exit

    @property
    def name(self) -> str:
        """The name of the logging context.

        Returns:
            The name of the logging context.
        """
        return self._name

    @classmethod
    def emit(
        cls,
        record: logging.LogRecord,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Emit a log record using the active logging context.

        This class method is called by stdout/stderr wrappers and logging
        handlers to route logs to the active log store.

        Args:
            record: The log record to emit.
            metadata: Additional metadata to attach to the log entry.
        """
        if context := LoggingContext.get():
            if context._disabled:
                return
            context._disabled = True
            try:
                message = record.getMessage()
                if message and message.strip():
                    if context._origin:
                        context._log_store.emit(
                            origin=context._origin,
                            record=record,
                            metadata=metadata,
                        )
            except Exception:
                logger.debug("Failed to emit log record", exc_info=True)
            finally:
                context._disabled = False

    def update(
        self,
        step_run: Optional["StepRunResponse"] = None,
        pipeline_run: Optional["PipelineRunResponse"] = None,
    ) -> None:
        """Update the metadata of the logging context.

        Args:
            step_run: The step run.
            pipeline_run: The pipeline run.
        """
        if step_run and self._origin:
            self._origin.metadata.update(
                get_step_log_metadata(step_run=step_run)
            )
        if pipeline_run and self._origin:
            self._origin.metadata.update(
                get_run_log_metadata(pipeline_run=pipeline_run)
            )

    def attach(
        self, response: Union["StepRunResponse", "PipelineRunResponse"]
    ) -> None:
        """Attach the logging context to a step run or pipeline run.

        Args:
            response: The step run or pipeline run.

        Raises:
            ValueError: If the response is not a step run or pipeline run.
        """
        if isinstance(response, StepRunResponse):
            log_update = LogsUpdate(step_run_id=response.id)
        elif isinstance(response, PipelineRunResponse):
            log_update = LogsUpdate(pipeline_run_id=response.id)
        else:
            raise ValueError("Response must be a step run or pipeline run.")

        Client().zen_store.update_logs(
            logs_id=self.log_model.id, logs_update=log_update
        )

    def __enter__(self) -> "LoggingContext":
        """Enter the context and set as active.

        Returns:
            self
        """
        with self._lock:
            super().__enter__()
            self._origin = self._log_store.register_origin(
                name=self.name,
                log_model=self.log_model,
                metadata=self._metadata,
            )

        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Exit the context and restore previous context.

        Args:
            exc_type: The class of the exception.
            exc_val: The instance of the exception.
            exc_tb: The traceback of the exception.
        """
        if exc_type is not None:
            LoggingContext.emit(
                logging.LogRecord(
                    name="",
                    level=logging.ERROR,
                    msg="An exception has occurred.",
                    args=(),
                    exc_info=(exc_type, exc_val, exc_tb) if exc_val else None,
                    func=None,
                    pathname="",
                    lineno=0,
                ),
                metadata={"zenml.event.type": "exception"},
            )

        with self._lock:
            super().__exit__(exc_type, exc_val, exc_tb)
            if self._origin:
                self._log_store.deregister_origin(
                    self._origin, blocking=self._block_on_exit
                )
                self._origin = None


def generate_logs_request(source: str) -> LogsRequest:
    """Generate a LogsRequest for logging.

    Args:
        source: The source of the logs (e.g., "client", "orchestrator", "step").

    Returns:
        A LogsRequest object.
    """
    from zenml.log_stores.artifact.artifact_log_store import (
        ArtifactLogStore,
        prepare_logs_uri,
    )

    client = Client()
    log_store = client.active_stack.log_store
    log_id = uuid4()

    if isinstance(log_store, ArtifactLogStore):
        artifact_store = client.active_stack.artifact_store
        return LogsRequest(
            id=log_id,
            project=client.active_project.id,
            source=source,
            uri=prepare_logs_uri(
                artifact_store=artifact_store,
                log_id=log_id,
            ),
            artifact_store_id=artifact_store.id,
        )
    else:
        return LogsRequest(
            id=log_id,
            project=client.active_project.id,
            source=source,
            log_store_id=log_store.id if log_store else None,
        )


def is_pipeline_logging_enabled(
    pipeline_configuration: PipelineConfiguration,
) -> bool:
    """Check if logging is enabled for a pipeline configuration.

    Args:
        pipeline_configuration: The pipeline configuration.

    Returns:
        True if logging is enabled, False if disabled.
    """
    if handle_bool_env_var(ENV_ZENML_DISABLE_PIPELINE_LOGS_STORAGE, False):
        return False
    elif pipeline_configuration.enable_pipeline_logs is not None:
        return pipeline_configuration.enable_pipeline_logs
    else:
        return True


def is_step_logging_enabled(
    step_configuration: StepConfiguration,
    pipeline_configuration: PipelineConfiguration,
) -> bool:
    """Check if logging is enabled for a step configuration.

    Args:
        step_configuration: The step configuration.
        pipeline_configuration: The pipeline configuration.

    Returns:
        True if logging is enabled, False if disabled.
    """
    from zenml.orchestrators.utils import is_setting_enabled

    if handle_bool_env_var(ENV_ZENML_DISABLE_STEP_LOGS_STORAGE, False):
        return False
    else:
        is_enabled_on_step = step_configuration.enable_step_logs
        is_enabled_on_pipeline = pipeline_configuration.enable_step_logs
        return is_setting_enabled(is_enabled_on_step, is_enabled_on_pipeline)


def search_logs_by_source(
    logs_collection: List[LogsResponse], source: str
) -> Optional[LogsResponse]:
    """Get the logs response for a given source.

    Args:
        logs_collection: The logs collection.
        source: The source of the logs.

    Returns:
        The logs response for the given source.
    """
    for log in logs_collection:
        if log.source == source:
            return log
    return None


def get_run_log_metadata(
    pipeline_run: "PipelineRunResponse",
) -> Dict[str, Any]:
    """Get the log metadata for a pipeline run.

    Args:
        pipeline_run: The pipeline run.

    Returns:
        The log metadata.
    """
    log_metadata = {
        "zenml.pipeline.run.id": str(pipeline_run.id),
        "zenml.pipeline.run.name": pipeline_run.name,
        "zenml.project.id": str(pipeline_run.project.id),
        "zenml.project.name": pipeline_run.project.name,
    }

    client = Client()
    server_info = client.zen_store.get_store_info()

    if server_info.is_pro_server():
        if server_info.pro_workspace_id:
            log_metadata.update(
                {
                    "zenml.workspace.id": str(server_info.pro_workspace_id),
                }
            )
        if server_info.pro_workspace_name:
            log_metadata.update(
                {
                    "zenml.workspace.name": server_info.pro_workspace_name,
                }
            )

    if pipeline_run.pipeline is not None:
        log_metadata.update(
            {
                "zenml.pipeline.id": str(pipeline_run.pipeline.id),
                "zenml.pipeline.name": pipeline_run.pipeline.name,
            }
        )

    if pipeline_run.stack is not None:
        log_metadata.update(
            {
                "zenml.stack.id": str(pipeline_run.stack.id),
                "zenml.stack.name": pipeline_run.stack.name,
            }
        )

    if pipeline_run.user is not None:
        log_metadata.update(
            {
                "zenml.user.id": str(pipeline_run.user.id),
                "zenml.user.name": pipeline_run.user.name,
            }
        )

    return log_metadata


def get_step_log_metadata(step_run: "StepRunResponse") -> Dict[str, Any]:
    """Get the log metadata for a step run.

    Args:
        step_run: The step run.

    Returns:
        The log metadata.
    """
    return {
        "zenml.step.run.id": str(step_run.id),
        "zenml.step.run.name": step_run.name,
    }


def fetch_logs(
    logs: "LogsResponse",
    zen_store: "BaseZenStore",
    limit: int,
) -> List["LogEntry"]:
    """Fetch logs from the log store.

    This function is designed to be called from the server side where we can't
    always instantiate the full Stack object due to missing integration dependencies.
    Instead, it directly instantiates the appropriate log store based on the logs model.

    Args:
        logs: The logs response model containing metadata about the logs.
        zen_store: The zen store instance.
        limit: Maximum number of log entries to return.

    Returns:
        List of log entries.

    Raises:
        DoesNotExistException: If the log store doesn't exist or is not the right type.
        NotImplementedError: If the log store's dependencies are not installed.
        RuntimeError: If the function is called from the client environment.
    """
    from zenml.log_stores.base_log_store import BaseLogStore
    from zenml.stack import StackComponent
    from zenml.zen_server.rbac.endpoint_utils import (
        verify_permissions_and_get_entity,
    )

    if ENV_ZENML_SERVER not in os.environ:
        # This utility function should not be called from the client environment
        # because it would cause instantiating the active log store again.
        raise RuntimeError(
            "This utility function is only supported in the server "
            "environment. Use the log store directly instead."
        )

    log_store: Optional[BaseLogStore] = None

    if logs.log_store_id:
        try:
            log_store_model = verify_permissions_and_get_entity(
                id=logs.log_store_id,
                get_method=zen_store.get_stack_component,
            )
        except KeyError:
            raise DoesNotExistException(
                f"Log store '{logs.log_store_id}' does not exist."
            )

        if not log_store_model.type == StackComponentType.LOG_STORE:
            raise DoesNotExistException(
                f"Stack component '{logs.log_store_id}' is not a log store."
            )

        try:
            log_store = cast(
                BaseLogStore,
                StackComponent.from_model(log_store_model),
            )
        except ImportError:
            raise NotImplementedError(
                f"Log store '{log_store_model.name}' could not be "
                "instantiated."
            )
    elif logs.artifact_store_id:
        from zenml.artifact_stores.base_artifact_store import BaseArtifactStore
        from zenml.log_stores.artifact.artifact_log_store import (
            ArtifactLogStore,
        )

        try:
            artifact_store_model = verify_permissions_and_get_entity(
                id=logs.artifact_store_id,
                get_method=zen_store.get_stack_component,
            )
        except KeyError:
            raise DoesNotExistException(
                f"Artifact store '{logs.artifact_store_id}' does not exist."
            )
        if not artifact_store_model.type == StackComponentType.ARTIFACT_STORE:
            raise DoesNotExistException(
                f"Stack component '{logs.artifact_store_id}' is not an artifact store."
            )
        artifact_store = cast(
            "BaseArtifactStore",
            StackComponent.from_model(artifact_store_model),
        )
        log_store = ArtifactLogStore.from_artifact_store(
            artifact_store=artifact_store
        )

    else:
        return []

    try:
        return log_store.fetch(logs_model=logs, limit=limit)
    finally:
        log_store.cleanup()


def setup_logging_context(
    source: str,
    pipeline_run: Optional["PipelineRunResponse"] = None,
    step_run: Optional["StepRunResponse"] = None,
    block_on_exit: bool = True,
) -> LoggingContext:
    """Setup a logging context for a given source.

    Args:
        source: The source of the logs.
        pipeline_run: The pipeline run.
        step_run: The step run.
        block_on_exit: Whether to block until all logs are flushed when the
            context is exited, if there are no more logging contexts active.

    Returns:
        A LoggingContext object.
    """
    log_metadata = {}

    logs_response: Optional[LogsResponse] = None

    if step_run and step_run.log_collection:
        logs_response = search_logs_by_source(
            logs_collection=step_run.log_collection,
            source=source,
        )
    elif pipeline_run and pipeline_run.log_collection:
        logs_response = search_logs_by_source(
            logs_collection=pipeline_run.log_collection,
            source=source,
        )

    logs_request = generate_logs_request(source=source)

    if step_run:
        log_metadata.update(get_step_log_metadata(step_run=step_run))
        logs_request.step_run_id = step_run.id

    if pipeline_run:
        log_metadata.update(get_run_log_metadata(pipeline_run=pipeline_run))
        if step_run is None:
            logs_request.pipeline_run_id = pipeline_run.id

    if logs_response is None:
        logs_response = Client().zen_store.create_logs(logs_request)

    return LoggingContext(
        name=str(logs_response.id),
        log_model=logs_response,
        block_on_exit=block_on_exit,
        **log_metadata,
    )



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""AWS Authentication Secret Schema definition."""

from typing import Optional

from pydantic import AliasChoices, Field

from zenml.secret.base_secret import BaseSecretSchema


class AWSSecretSchema(BaseSecretSchema):
    """AWS Authentication Secret Schema definition.

    This schema supports both AWS-prefixed field names (for AWS S3) and
    generic field names (for S3-compatible storage like MinIO, Alibaba OSS).

    Supported field names:
        - aws_access_key_id / access_key_id
        - aws_secret_access_key / secret_access_key
        - aws_session_token / session_token
    """

    aws_access_key_id: str = Field(
        validation_alias=AliasChoices("aws_access_key_id", "access_key_id"),
    )
    aws_secret_access_key: str = Field(
        validation_alias=AliasChoices(
            "aws_secret_access_key", "secret_access_key"
        ),
    )
    aws_session_token: Optional[str] = Field(
        default=None,
        validation_alias=AliasChoices("aws_session_token", "session_token"),
    )



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the ZenML HyperAI orchestrator."""

import os
import re
import tempfile
from shlex import quote
from typing import IO, TYPE_CHECKING, Any, Dict, Optional, Type, cast

import paramiko
import yaml

from zenml.config.base_settings import BaseSettings
from zenml.entrypoints import StepEntrypointConfiguration
from zenml.enums import StackComponentType
from zenml.integrations.hyperai.flavors.hyperai_orchestrator_flavor import (
    HyperAIOrchestratorConfig,
    HyperAIOrchestratorSettings,
)
from zenml.logger import get_logger
from zenml.orchestrators import ContainerizedOrchestrator, SubmissionResult
from zenml.stack import Stack, StackValidator

if TYPE_CHECKING:
    from zenml.models import PipelineRunResponse, PipelineSnapshotResponse

logger = get_logger(__name__)

ENV_ZENML_HYPERAI_RUN_ID = "ZENML_HYPERAI_ORCHESTRATOR_RUN_ID"


class HyperAIOrchestrator(ContainerizedOrchestrator):
    """Orchestrator responsible for running pipelines on HyperAI instances."""

    @property
    def config(self) -> HyperAIOrchestratorConfig:
        """Returns the `HyperAIOrchestratorConfig` config.

        Returns:
            The configuration.
        """
        return cast(HyperAIOrchestratorConfig, self._config)

    @property
    def settings_class(self) -> Optional[Type["BaseSettings"]]:
        """Settings class for the HyperAI orchestrator.

        Returns:
            The settings class.
        """
        return HyperAIOrchestratorSettings

    @property
    def validator(self) -> Optional[StackValidator]:
        """Ensures there is an image builder in the stack.

        Returns:
            A `StackValidator` instance.
        """
        return StackValidator(
            required_components={
                StackComponentType.CONTAINER_REGISTRY,
                StackComponentType.IMAGE_BUILDER,
            }
        )

    def get_orchestrator_run_id(self) -> str:
        """Returns the active orchestrator run id.

        Raises:
            RuntimeError: If the environment variable specifying the run id
                is not set.

        Returns:
            The orchestrator run id.
        """
        try:
            return os.environ[ENV_ZENML_HYPERAI_RUN_ID]
        except KeyError:
            raise RuntimeError(
                "Unable to read run id from environment variable "
                f"{ENV_ZENML_HYPERAI_RUN_ID}."
            )

    def _validate_mount_path(self, path: str) -> str:
        """Validates if a given string is in a valid path format.

        Args:
            path: The path to be validated.

        Returns:
            The path in a valid format.

        Raises:
            RuntimeError: If the path is not in a valid format.
        """
        # Define a regular expression pattern to match a valid path format
        pattern = r'^(?:[a-zA-Z]:\\(\\[^\\/:*?"<>|]*)*$|^/([^\0]*)*$)'

        if bool(re.match(pattern, path)):
            return path
        else:
            raise RuntimeError(
                f"Path '{path}' is not in a valid format, so a mount cannot be established."
            )

    def _escape_shell_command(self, command: str) -> str:
        """Escapes a shell command.

        Args:
            command: The command to escape.

        Returns:
            The escaped command.
        """
        return quote(command)

    def _scp_to_hyperai_instance(
        self,
        paramiko_client: paramiko.SSHClient,
        f: IO[str],
        directory_name: str,
        file_name: str,
        description: str,
    ) -> None:
        """Copies a file to a HyperAI instance using SCP.

        Args:
            paramiko_client: The SSH client to use for the SCP transfer.
            f: The file to transfer.
            directory_name: The directory on the HyperAI instance to transfer
                the file to.
            file_name: The name of the file being transferred.
            description: A description of the file being transferred.

        Raises:
            RuntimeError: If the file cannot be written to the HyperAI instance.
        """
        try:
            scp_client = paramiko_client.open_sftp()
            scp_client.put(f.name, f"{directory_name}/{file_name}")
            scp_client.close()
        except FileNotFoundError:
            raise RuntimeError(
                f"Failed to write {description} to HyperAI instance. Does the user have permissions to write?"
            )

    def submit_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        base_environment: Dict[str, str],
        step_environments: Dict[str, Dict[str, str]],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits a pipeline to the orchestrator.

        This method should only submit the pipeline and not wait for it to
        complete. If the orchestrator is configured to wait for the pipeline run
        to complete, a function that waits for the pipeline run to complete can
        be passed as part of the submission result.

        Assumes that:
        - A HyperAI (hyperai.ai) instance is running on the configured IP address.
        - The HyperAI instance has been configured to allow SSH connections from the
            machine running the pipeline.
        - Docker and Docker Compose are installed on the HyperAI instance.
        - A key pair has been generated and the public key has been added to the
            HyperAI instance's `authorized_keys` file.
        - The private key is available in a HyperAI service connector linked to this
            orchestrator.

        Args:
            snapshot: The pipeline snapshot to submit.
            stack: The stack the pipeline will run on.
            base_environment: Base environment shared by all steps. This should
                be set if your orchestrator for example runs one container that
                is responsible for starting all the steps.
            step_environments: Environment variables to set when executing
                specific steps.
            placeholder_run: An optional placeholder run for the snapshot.

        Raises:
            RuntimeError: If running the pipeline fails.

        Returns:
            Optional submission result.
        """
        from zenml.integrations.hyperai.service_connectors.hyperai_service_connector import (
            HyperAIServiceConnector,
        )

        compose_definition: Dict[str, Any] = {"version": "3", "services": {}}
        snapshot_id = snapshot.id

        os.environ[ENV_ZENML_HYPERAI_RUN_ID] = str(snapshot_id)

        # Add each step as a service to the Docker Compose definition
        logger.info("Preparing pipeline steps for deployment.")
        for step_name, step in snapshot.step_configurations.items():
            # Get image
            image = self.get_image(snapshot=snapshot, step_name=step_name)

            step_settings = cast(
                HyperAIOrchestratorSettings, self.get_settings(step)
            )
            container_name = f"{snapshot_id}-{step_name}"

            # Make Compose service definition for step
            compose_definition["services"][container_name] = {
                "image": image,
                "container_name": container_name,
                "network_mode": "host",
                "entrypoint": StepEntrypointConfiguration.get_entrypoint_command(),
                "command": StepEntrypointConfiguration.get_entrypoint_arguments(
                    step_name=step_name, snapshot_id=snapshot.id
                ),
                "volumes": [
                    "{}:{}".format(
                        self._validate_mount_path(mount_from),
                        self._validate_mount_path(mount_to),
                    )
                    for mount_from, mount_to in step_settings.mounts_from_to.items()
                ],
            }

            # Depending on GPU setting, add GPU support to service definition
            if self.config.gpu_enabled_in_container:
                compose_definition["services"][container_name]["deploy"] = {
                    "resources": {
                        "reservations": {
                            "devices": [
                                {"driver": "nvidia", "capabilities": ["gpu"]}
                            ]
                        }
                    }
                }

            # Add run ID to step environment variables
            step_env = step_environments[step_name].copy()

            # Depending on whether it is a scheduled or a realtime pipeline, add
            # potential .env file to service definition for deployment ID override.
            if snapshot.schedule:
                compose_definition["services"][container_name]["env_file"] = [
                    ".env"
                ]
            else:
                step_env[ENV_ZENML_HYPERAI_RUN_ID] = str(snapshot_id)

            compose_definition["services"][container_name]["environment"] = (
                step_env
            )

            # Add dependency on upstream steps if applicable
            upstream_steps = step.spec.upstream_steps

            if len(upstream_steps) > 0:
                compose_definition["services"][container_name][
                    "depends_on"
                ] = {}

                for upstream_step_name in upstream_steps:
                    upstream_container_name = (
                        f"{snapshot_id}-{upstream_step_name}"
                    )
                    compose_definition["services"][container_name][
                        "depends_on"
                    ].update(
                        {
                            upstream_container_name: {
                                "condition": "service_completed_successfully"
                            }
                        }
                    )

        compose_definition_yaml: str = yaml.dump(compose_definition)

        # Connect to configured HyperAI instance
        logger.info(
            "Connecting to HyperAI instance and placing Docker Compose file."
        )
        if connector := self.get_connector():
            paramiko_client = connector.connect()
            if paramiko_client is None:
                raise RuntimeError(
                    "Expected to receive a `paramiko.SSHClient` object from the "
                    "linked connector, but got `None`. This likely originates from "
                    "a misconfigured service connector, typically caused by a wrong "
                    "SSH key type being selected. Please check your "
                    "`hyperai_orchestrator` configuration and make sure that the "
                    "`ssh_key_type` of its connected service connector is set to the "
                    "correct value."
                )
            elif not isinstance(paramiko_client, paramiko.SSHClient):
                raise RuntimeError(
                    f"Expected to receive a `paramiko.SSHClient` object from the "
                    f"linked connector, but got type `{type(paramiko_client)}`."
                )
        else:
            raise RuntimeError(
                "You must link a HyperAI service connector to the orchestrator."
            )

        # Get container registry autologin setting
        if self.config.container_registry_autologin:
            logger.info(
                "Attempting to automatically log in to container registry used by stack."
            )

            # Select stack container registry
            container_registry = stack.container_registry

            # Raise error if no container registry is found
            if not container_registry:
                raise RuntimeError(
                    "Unable to find container registry in stack."
                )

            # Get container registry credentials from its config
            credentials = container_registry.credentials
            if credentials is None:
                raise RuntimeError(
                    "The container registry in the active stack has no "
                    "credentials or service connector configured, but the "
                    "HyperAI orchestrator is set to autologin to the container "
                    "registry. Please configure the container registry with "
                    "credentials or turn off the `container_registry_autologin` "
                    "setting in the HyperAI orchestrator configuration."
                )

            container_registry_url = container_registry.config.uri
            (
                container_registry_username,
                container_registry_password,
            ) = credentials

            # Escape inputs
            container_registry_username = self._escape_shell_command(
                container_registry_username
            )
            container_registry_url = self._escape_shell_command(
                container_registry_url
            )

            # Log in to container registry using --password-stdin
            stdin, stdout, stderr = paramiko_client.exec_command(  # nosec
                f"docker login -u {container_registry_username} "
                f"--password-stdin {container_registry_url}"
            )
            # Send the password to stdin
            stdin.channel.send(
                f"{container_registry_password}\n".encode("utf-8")
            )
            stdin.channel.shutdown_write()

            # Log stdout
            for line in stdout.readlines():
                logger.info(line)

        # Get username from connector
        assert isinstance(connector, HyperAIServiceConnector)
        username = connector.config.username

        # Set up pipeline-runs directory if it doesn't exist
        nonscheduled_directory_name = self._escape_shell_command(
            f"/home/{username}/pipeline-runs"
        )
        directory_name = (
            nonscheduled_directory_name
            if not snapshot.schedule
            else self._escape_shell_command(
                f"/home/{username}/scheduled-pipeline-runs"
            )
        )
        stdin, stdout, stderr = paramiko_client.exec_command(  # nosec
            f"mkdir -p {directory_name}"
        )

        # Get pipeline run id and create directory for it
        orchestrator_run_id = self.get_orchestrator_run_id()
        directory_name = self._escape_shell_command(
            f"{directory_name}/{orchestrator_run_id}"
        )
        stdin, stdout, stderr = paramiko_client.exec_command(  # nosec
            f"mkdir -p {directory_name}"
        )

        # Remove all folders from nonscheduled pipelines if they are 7 days old or older
        if self.config.automatic_cleanup_pipeline_files:
            logger.info(
                "Cleaning up old pipeline files on HyperAI instance. This may take a while."
            )
            stdin, stdout, stderr = paramiko_client.exec_command(  # nosec
                f"find {nonscheduled_directory_name} -type d -ctime +7 -exec rm -rf {{}} +"
            )

        # Create temporary file and write Docker Compose file to it
        with tempfile.NamedTemporaryFile(mode="w", delete=True) as f:
            # Write Docker Compose file to temporary file
            with f.file as f_:
                f_.write(compose_definition_yaml)

            # Scp Docker Compose file to HyperAI instance
            self._scp_to_hyperai_instance(
                paramiko_client,
                f,
                directory_name,
                file_name="docker-compose.yml",
                description="Docker Compose file",
            )

        # Create temporary file and write script to it
        with tempfile.NamedTemporaryFile(mode="w", delete=True) as f:
            # Define bash line and command line
            bash_line = "#!/bin/bash\n"
            command_line = rf'cd {directory_name} && echo {ENV_ZENML_HYPERAI_RUN_ID}="{snapshot_id}_$(date +\%s)" > .env && docker compose up -d'

            # Write script to temporary file
            with f.file as f_:
                f_.write(bash_line)
                f_.write(command_line)

            # Scp script to HyperAI instance
            self._scp_to_hyperai_instance(
                paramiko_client,
                f,
                directory_name,
                file_name="run_pipeline.sh",
                description="startup script",
            )

        # Run or schedule Docker Compose file depending on settings
        if not snapshot.schedule:
            logger.info(
                "Starting ZenML pipeline on HyperAI instance. Depending on the size of your container image, this may take a while..."
            )
            stdin, stdout, stderr = paramiko_client.exec_command(  # nosec
                f"cd {directory_name} && docker compose up -d"
            )

            # Log errors in case of failure
            for line in stderr.readlines():
                logger.info(line)
        elif snapshot.schedule and snapshot.schedule.cron_expression:
            # Get cron expression for scheduled pipeline
            cron_expression = snapshot.schedule.cron_expression
            if not cron_expression:
                raise RuntimeError(
                    "A cron expression is required for scheduled pipelines."
                )
            expected_cron_pattern = r"^(?:(?:[0-9]|[1-5][0-9]|60)(?:,(?:[0-9]|[1-5][0-9]|60))*|[*](?:\/[1-9][0-9]*)?)(?:[ \t]+(?:(?:[0-9]|[0-5][0-9]|60)(?:,(?:[0-9]|[0-5][0-9]|60))*|[*](?:\/[1-9][0-9]*)?)){4}$"
            if not re.match(expected_cron_pattern, cron_expression):
                raise RuntimeError(
                    f"The cron expression '{cron_expression}' is not in a valid format."
                )

            # Log about scheduling
            logger.info(f"Requested cron expression: {cron_expression}")
            logger.info("Scheduling ZenML pipeline on HyperAI instance...")

            # Create cron job for scheduled pipeline on HyperAI instance
            stdin, stdout, stderr = paramiko_client.exec_command(  # nosec
                f"(crontab -l ; echo '{cron_expression} bash {directory_name}/run_pipeline.sh') | crontab -"
            )

            logger.info(
                f"Pipeline scheduled successfully in crontab with cron expression: {cron_expression}"
            )
        elif snapshot.schedule and snapshot.schedule.run_once_start_time:
            # Get start time for scheduled pipeline
            start_time = snapshot.schedule.run_once_start_time

            # Log about scheduling
            logger.info(f"Requested start time: {start_time}")
            logger.info("Scheduling ZenML pipeline on HyperAI instance...")

            # Check if `at` is installed on HyperAI instance
            stdin, stdout, stderr = paramiko_client.exec_command(  # nosec
                "which at"
            )
            if not stdout.readlines():
                raise RuntimeError(
                    "The `at` command is not installed on the HyperAI instance. Please install it to use start times for scheduled pipelines."
                )

            # Convert start time into YYYYMMDDHHMM.SS format
            start_time_str = start_time.strftime("%Y%m%d%H%M.%S")

            # Create cron job for scheduled pipeline on HyperAI instance
            stdin, stdout, stderr = paramiko_client.exec_command(  # nosec
                f"echo 'bash {directory_name}/run_pipeline.sh' | at -t {start_time_str}"
            )

            logger.info(
                f"Pipeline scheduled successfully to run once at: {start_time}"
            )
        else:
            raise RuntimeError(
                "A cron expression or start time is required for scheduled pipelines."
            )

        return None



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""ZenML pipeline decorator definition."""

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    TypeVar,
    Union,
    overload,
)
from uuid import UUID

from zenml.enums import ExecutionMode
from zenml.logger import get_logger

if TYPE_CHECKING:
    from zenml.config.base_settings import SettingsOrDict
    from zenml.config.cache_policy import CachePolicyOrString
    from zenml.config.retry_config import StepRetryConfig
    from zenml.model.model import Model
    from zenml.pipelines.pipeline_definition import Pipeline
    from zenml.steps.base_step import BaseStep
    from zenml.types import HookSpecification, InitHookSpecification
    from zenml.utils.tag_utils import Tag

    F = TypeVar("F", bound=Callable[..., Any])

logger = get_logger(__name__)


@overload
def pipeline(_func: "F") -> "Pipeline": ...


@overload
def pipeline(
    *,
    name: Optional[str] = None,
    dynamic: Optional[bool] = None,
    depends_on: Optional[List["BaseStep"]] = None,
    enable_cache: Optional[bool] = None,
    enable_artifact_metadata: Optional[bool] = None,
    enable_step_logs: Optional[bool] = None,
    enable_heartbeat: Optional[bool] = None,
    environment: Optional[Dict[str, Any]] = None,
    secrets: Optional[List[Union[UUID, str]]] = None,
    enable_pipeline_logs: Optional[bool] = None,
    settings: Optional[Dict[str, "SettingsOrDict"]] = None,
    tags: Optional[List[Union[str, "Tag"]]] = None,
    extra: Optional[Dict[str, Any]] = None,
    on_failure: Optional["HookSpecification"] = None,
    on_success: Optional["HookSpecification"] = None,
    on_init: Optional["InitHookSpecification"] = None,
    on_init_kwargs: Optional[Dict[str, Any]] = None,
    on_cleanup: Optional["HookSpecification"] = None,
    model: Optional["Model"] = None,
    retry: Optional["StepRetryConfig"] = None,
    substitutions: Optional[Dict[str, str]] = None,
    execution_mode: Optional["ExecutionMode"] = None,
    cache_policy: Optional["CachePolicyOrString"] = None,
) -> Callable[["F"], "Pipeline"]: ...


def pipeline(
    _func: Optional["F"] = None,
    *,
    name: Optional[str] = None,
    dynamic: Optional[bool] = None,
    depends_on: Optional[List["BaseStep"]] = None,
    enable_cache: Optional[bool] = None,
    enable_artifact_metadata: Optional[bool] = None,
    enable_step_logs: Optional[bool] = None,
    enable_heartbeat: Optional[bool] = None,
    environment: Optional[Dict[str, Any]] = None,
    secrets: Optional[List[Union[UUID, str]]] = None,
    enable_pipeline_logs: Optional[bool] = None,
    settings: Optional[Dict[str, "SettingsOrDict"]] = None,
    tags: Optional[List[Union[str, "Tag"]]] = None,
    extra: Optional[Dict[str, Any]] = None,
    on_failure: Optional["HookSpecification"] = None,
    on_success: Optional["HookSpecification"] = None,
    on_init: Optional["InitHookSpecification"] = None,
    on_init_kwargs: Optional[Dict[str, Any]] = None,
    on_cleanup: Optional["HookSpecification"] = None,
    model: Optional["Model"] = None,
    retry: Optional["StepRetryConfig"] = None,
    substitutions: Optional[Dict[str, str]] = None,
    execution_mode: Optional["ExecutionMode"] = None,
    cache_policy: Optional["CachePolicyOrString"] = None,
) -> Union["Pipeline", Callable[["F"], "Pipeline"]]:
    """Decorator to create a pipeline.

    Args:
        _func: The decorated function.
        name: The name of the pipeline. If left empty, the name of the
            decorated function will be used as a fallback.
        dynamic: Whether this is a dynamic pipeline or not.
        depends_on: The steps that this pipeline depends on.
        enable_cache: Whether to use caching or not.
        enable_artifact_metadata: Whether to enable artifact metadata or not.
        enable_step_logs: If step logs should be enabled for this pipeline.
        enable_heartbeat: If heartbeat should be enabled for this pipeline.
        environment: Environment variables to set when running this pipeline.
        secrets: Secrets to set as environment variables when running this
            pipeline.
        enable_pipeline_logs: If pipeline logs should be enabled for this pipeline.
        settings: Settings for this pipeline.
        tags: Tags to apply to runs of the pipeline.
        extra: Extra configurations for this pipeline.
        on_failure: Callback function in event of failure of the step. Can be a
            function with a single argument of type `BaseException`, or a source
            path to such a function (e.g. `module.my_function`).
        on_success: Callback function in event of success of the step. Can be a
            function with no arguments, or a source path to such a function
            (e.g. `module.my_function`).
        on_init: Callback function to run on initialization of the pipeline. Can
            be a function with no arguments, or a source path to such a function
            (e.g. `module.my_function`) if the function returns a value, it will
            be stored as the pipeline state.
        on_init_kwargs: Arguments for the init hook.
        on_cleanup: Callback function to run on cleanup of the pipeline. Can be a
            function with no arguments, or a source path to such a function
            (e.g. `module.my_function`).
        model: configuration of the model in the Model Control Plane.
        retry: Retry configuration for the pipeline steps.
        substitutions: Extra substitutions for pipeline run, model and
            artifact name placeholders.
        execution_mode: The execution mode to use for the pipeline.
        cache_policy: Cache policy for this pipeline.

    Returns:
        A pipeline instance.
    """

    def inner_decorator(func: "F") -> "Pipeline":
        from zenml.pipelines.pipeline_definition import Pipeline

        PipelineClass = Pipeline
        pipeline_args: Dict[str, Any] = {}

        if dynamic:
            from zenml.pipelines.dynamic.pipeline_definition import (
                DynamicPipeline,
            )

            PipelineClass = DynamicPipeline

            pipeline_args = {
                "depends_on": depends_on,
            }
        elif depends_on:
            logger.warning(
                "The `depends_on` argument is not supported "
                "for static pipelines and will be ignored."
            )

        p = PipelineClass(
            name=name or func.__name__,
            entrypoint=func,
            enable_cache=enable_cache,
            enable_artifact_metadata=enable_artifact_metadata,
            enable_step_logs=enable_step_logs,
            enable_heartbeat=enable_heartbeat,
            environment=environment,
            secrets=secrets,
            enable_pipeline_logs=enable_pipeline_logs,
            settings=settings,
            tags=tags,
            extra=extra,
            on_failure=on_failure,
            on_success=on_success,
            on_init=on_init,
            on_init_kwargs=on_init_kwargs,
            on_cleanup=on_cleanup,
            model=model,
            retry=retry,
            substitutions=substitutions,
            execution_mode=execution_mode,
            cache_policy=cache_policy,
            **pipeline_args,
        )

        p.__doc__ = func.__doc__
        return p

    return inner_decorator if _func is None else inner_decorator(_func)



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""GCP integration flavors."""

from zenml.integrations.gcp.flavors.gcp_artifact_store_flavor import (
    GCPArtifactStoreConfig,
    GCPArtifactStoreFlavor,
)
from zenml.integrations.gcp.flavors.gcp_deployer_flavor import (
    GCPDeployerConfig,
    GCPDeployerFlavor,
)
from zenml.integrations.gcp.flavors.gcp_image_builder_flavor import (
    GCPImageBuilderConfig,
    GCPImageBuilderFlavor,
)
from zenml.integrations.gcp.flavors.vertex_experiment_tracker_flavor import (
    VertexExperimentTrackerConfig,
    VertexExperimentTrackerFlavor,
)
from zenml.integrations.gcp.flavors.vertex_orchestrator_flavor import (
    VertexOrchestratorConfig,
    VertexOrchestratorFlavor,
)
from zenml.integrations.gcp.flavors.vertex_step_operator_flavor import (
    VertexStepOperatorConfig,
    VertexStepOperatorFlavor,
)

__all__ = [
    "GCPArtifactStoreFlavor",
    "GCPArtifactStoreConfig",
    "GCPDeployerFlavor",
    "GCPDeployerConfig",
    "GCPImageBuilderFlavor",
    "GCPImageBuilderConfig",
    "VertexExperimentTrackerFlavor",
    "VertexExperimentTrackerConfig",
    "VertexOrchestratorFlavor",
    "VertexOrchestratorConfig",
    "VertexStepOperatorFlavor",
    "VertexStepOperatorConfig",
]



"""add log stores [5c0a1c787128].

Revision ID: 5c0a1c787128
Revises: 0.91.2
Create Date: 2025-10-24 10:06:54.402219

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "5c0a1c787128"
down_revision = "0.92.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    with op.batch_alter_table("logs", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "log_store_id", sqlmodel.sql.sqltypes.GUID(), nullable=True
            )
        )
        batch_op.alter_column("uri", existing_type=sa.TEXT(), nullable=True)
        batch_op.alter_column(
            "artifact_store_id",
            existing_type=sa.CHAR(length=32),
            nullable=True,
        )
        batch_op.create_foreign_key(
            "fk_logs_log_store_id_stack_component",
            "stack_component",
            ["log_store_id"],
            ["id"],
            ondelete="CASCADE",
        )


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    with op.batch_alter_table("logs", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_logs_log_store_id_stack_component", type_="foreignkey"
        )
        batch_op.alter_column(
            "artifact_store_id",
            existing_type=sa.CHAR(length=32),
            nullable=False,
        )
        batch_op.alter_column("uri", existing_type=sa.TEXT(), nullable=False)
        batch_op.drop_column("log_store_id")



#  Copyright (c) ZenML GmbH 2026. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Endpoint definitions for run wait conditions."""

from uuid import UUID

from fastapi import APIRouter, Depends, Security

from zenml.constants import (
    API,
    RESOLVE,
    RUN_WAIT_CONDITIONS,
    VERSION_1,
)
from zenml.enums import RunWaitConditionStatus
from zenml.models import (
    Page,
    RunWaitConditionFilter,
    RunWaitConditionLeaseUpdate,
    RunWaitConditionRequest,
    RunWaitConditionResolveRequest,
    RunWaitConditionResponse,
)
from zenml.zen_server.auth import AuthContext, authorize
from zenml.zen_server.exceptions import error_response
from zenml.zen_server.rbac.endpoint_utils import (
    verify_permissions_and_create_entity,
)
from zenml.zen_server.rbac.models import Action, ResourceType
from zenml.zen_server.rbac.utils import (
    dehydrate_page,
    dehydrate_response_model,
    get_allowed_resource_ids,
    verify_permission_for_model,
)
from zenml.zen_server.utils import (
    async_fastapi_endpoint_wrapper,
    make_dependable,
    set_filter_project_scope,
    zen_store,
)

router = APIRouter(
    prefix=API + VERSION_1,
    tags=["run_wait_conditions"],
    responses={401: error_response, 403: error_response},
)


@router.post(
    RUN_WAIT_CONDITIONS,
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def create_run_wait_condition(
    run_wait_condition: RunWaitConditionRequest,
    _: AuthContext = Security(authorize),
) -> RunWaitConditionResponse:
    """Create a wait condition for a run.

    Args:
        run_wait_condition: Wait condition creation payload.

    Returns:
        The created wait condition.
    """
    run = zen_store().get_run(run_wait_condition.run, hydrate=False)
    return verify_permissions_and_create_entity(
        request_model=run_wait_condition,
        create_method=zen_store().create_run_wait_condition,
        surrogate_models=[run],
    )


@router.get(
    RUN_WAIT_CONDITIONS,
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper(deduplicate=True)
def list_wait_conditions(
    run_wait_condition_filter_model: RunWaitConditionFilter = Depends(
        make_dependable(RunWaitConditionFilter)
    ),
    hydrate: bool = False,
    auth_context: AuthContext = Security(authorize),
) -> Page[RunWaitConditionResponse]:
    """List wait conditions across runs.

    Args:
        run_wait_condition_filter_model: Filter model.
        hydrate: Whether to hydrate metadata/resources.
        auth_context: Request auth context.

    Returns:
        A page of wait conditions.
    """
    set_filter_project_scope(run_wait_condition_filter_model)
    assert isinstance(run_wait_condition_filter_model.project, UUID)

    allowed_pipeline_run_ids = get_allowed_resource_ids(
        resource_type=ResourceType.PIPELINE_RUN,
        project_id=run_wait_condition_filter_model.project,
    )
    run_wait_condition_filter_model.configure_rbac(
        authenticated_user_id=auth_context.user.id,
        run_id=allowed_pipeline_run_ids,
    )

    page = zen_store().list_run_wait_conditions(
        run_wait_condition_filter_model=run_wait_condition_filter_model,
        hydrate=hydrate,
    )
    return dehydrate_page(page)


@router.get(
    RUN_WAIT_CONDITIONS + "/{run_wait_condition_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper(deduplicate=True)
def get_run_wait_condition(
    run_wait_condition_id: UUID,
    hydrate: bool = True,
    _: AuthContext = Security(authorize),
) -> RunWaitConditionResponse:
    """Get a wait condition by ID.

    Args:
        run_wait_condition_id: Wait condition ID.
        hydrate: Whether to hydrate metadata/resources.

    Returns:
        The requested wait condition.
    """
    condition = zen_store().get_run_wait_condition(
        run_wait_condition_id=run_wait_condition_id, hydrate=hydrate
    )
    verify_permission_for_model(condition.run, action=Action.READ)
    return dehydrate_response_model(condition)


@router.put(
    RUN_WAIT_CONDITIONS + "/{run_wait_condition_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def update_run_wait_condition_lease(
    run_wait_condition_id: UUID,
    lease_update: RunWaitConditionLeaseUpdate,
    _: AuthContext = Security(authorize),
) -> RunWaitConditionStatus:
    """Update a wait condition polling lease.

    Args:
        run_wait_condition_id: Wait condition ID.
        lease_update: Lease refresh payload.

    Returns:
        The current wait condition status after attempting the lease update.
    """
    condition = zen_store().get_run_wait_condition(
        run_wait_condition_id=run_wait_condition_id,
        hydrate=False,
    )
    verify_permission_for_model(model=condition.run, action=Action.UPDATE)
    return zen_store().update_run_wait_condition_lease(
        run_wait_condition_id=run_wait_condition_id,
        lease_update=lease_update,
    )


@router.put(
    RUN_WAIT_CONDITIONS + "/{run_wait_condition_id}" + RESOLVE,
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def resolve_run_wait_condition(
    run_wait_condition_id: UUID,
    resolve_request: RunWaitConditionResolveRequest,
    _: AuthContext = Security(authorize),
) -> RunWaitConditionResponse:
    """Resolve a run wait condition.

    Args:
        run_wait_condition_id: Wait condition ID.
        resolve_request: Resolution payload.

    Returns:
        The resolved wait condition.
    """
    condition = zen_store().get_run_wait_condition(
        run_wait_condition_id=run_wait_condition_id,
        hydrate=False,
    )
    verify_permission_for_model(model=condition.run, action=Action.UPDATE)

    resolved_condition = zen_store().resolve_run_wait_condition(
        run_wait_condition_id=run_wait_condition_id,
        resolve_request=resolve_request,
    )
    return resolved_condition
